/**
 * VEditor.
 *
 * @author dongyado
 */
class VEditor {

    constructor(options) {
        let EDITOR   = this;
        this.options = {};

        // 校验数据
        if (!options.container_id) {
            throw new Error("container_id required");
        }

        this.container = $("#" + options.container_id);
        if (this.container.length === 0) {
            throw new Error("Invalid container_id : #" + container_id);
        }

        this.options.maxEditorWidth = options.containerWidth;
        this.options.containerWidth = !options.containerWidth ? this.container.width() : options.containerWidth;
        this.options.containerHeight = !options.containerHeight ? this.container.height() : options.containerHeight;

        this.container.html("");
        this.container.css({
            width: this.options.containerWidth + "px",
            height: this.options.containerHeight + "px",
            position: "relative",
            overflow: "hidden"
        });

        // 初始化常用的 html 变量
        this.initHtml();

        this.widgets       = {}; // 已经创建的组件
        this.currentAssets = {}; // 当前素材配置
        this.currentOrders = []; // 素材顺序
        this.currentGroups = {}; // 当前 groups
        this.groupWidgets = {};  // 存储组 widgets
        this.groupBoxes   = {};  // 存储组 wrapper box
        this.croppedData   = {}; // 完成截图的数据
        this.croppedCount  = 0;  // 截图计数
        this.assetSettings = {}; // 素材特殊设置
        this.editedAssets  = {}; // 素材是否编辑标记 map, 已经编辑过的素材将会被设置为 true

        if (options.assetSettings) {
            this.assetSettings = options.assetSettings;
        }

        // 调试配置
        this.debug            = false;
        this.enable_debug_box = false;

        // append debug box
        if (options.enable_debug_box) {
            this.enable_debug_box = true;
            this.debug_box        = $(this.htmls.debug_box_html);
            this.debug_count      = 0;
            $("body").append(this.debug_box);
        }

        if (options.debug) {
            this.debug = options.debug;
        }

        // 初始化模板配置文件
        if (options.config_json) {
            this.parseConfigJson(options.config_json);
        }


        // 当前组件和组件 id
        EDITOR.currentWidget = null;
        EDITOR.currentId     = null;


        // 控件选择监听器
        let chooseWidgetListener = function (e) {
            let mousePos = {
                x: e.originalEvent.clientX,
                y: e.originalEvent.clientY
            };

            EDITOR.chooseWidget(mousePos);

            if (EDITOR.currentWidget) {
                EDITOR.currentWidget.click(e);
            }
        };


        // 为编辑器容器清除和重新绑定监听器
        this.container.unbind("click");
        this.container.bind("click", chooseWidgetListener);

        // 绑定事件
        this.bindEvents();
    }


    /**
     * 绑定控件编辑事件
     *
     * */
    bindEvents() {
        let EDITOR  = this;
        // 检查当前组件是否可用，是否可移动
        let canMove = function () {
            if (EDITOR.currentWidget === null || !EDITOR.currentWidget.moveable) {
                return false;
            }
            return true;
        };


        // 编辑事件绑定
        let hammertime = new Hammer(this.container.get(0));
        hammertime.get('rotate').set({enable: true});
        hammertime.get('pinch').set({enable: true});

        hammertime.get('pan').set({direction: Hammer.DIRECTION_ALL});

        // 绑定旋转事件
        hammertime.on("rotatestart", function (e) {
            let mousePos = {
                x: e.center.x,
                y: e.center.y
            };
            EDITOR.chooseWidget(mousePos);

            if (!canMove()) return;

            EDITOR.currentWidget.moveData.rotation = e.rotation;
            EDITOR.log(EDITOR.currentWidget.cropper.getStyle());
        });

        hammertime.on("rotate", function (e) {
                if (!canMove() || EDITOR.currentWidget.cropper === undefined) return;

                if (e.rotation === 0) return;

                let deg    = e.rotation - EDITOR.currentWidget.moveData.rotation;
                let anchor = {x: e.center.x, y: e.center.y};
                EDITOR.currentWidget.cropper.rotate(deg, anchor);
                EDITOR.currentWidget.moveData.rotation = e.rotation;
            }
        );

        hammertime.on("rotateend", function (e) {
            if (!canMove()) return;

            EDITOR.currentWidget.moveData.rotation = e.rotation;
            EDITOR.doImageCrop(EDITOR.currentId, EDITOR.currentWidget);
            EDITOR.log(EDITOR.currentWidget.cropper.getStyle());
        });


        // 缩放
        hammertime.on("pinchstart", function (e) {
                let mousePos = {
                    x: e.center.x,
                    y: e.center.y
                };
                EDITOR.chooseWidget(mousePos);

                if (!canMove()) return;

                EDITOR.currentWidget.moveData.zoomScale = e.scale;
            }
        );

        hammertime.on("pinch", function (e) {
                if (!canMove() || EDITOR.currentWidget.cropper === undefined) return;

                let zoom   = e.scale / EDITOR.currentWidget.moveData.zoomScale;
                let anchor = {x: e.center.x, y: e.center.y};

                EDITOR.currentWidget.cropper.zoom(zoom, anchor);
                EDITOR.currentWidget.moveData.zoomScale = e.scale;
            }
        );

        hammertime.on("pinchend", function () {
                if (!canMove()) return;

                EDITOR.doImageCrop(EDITOR.currentId, EDITOR.currentWidget);
            }
        );


        // 移动
        hammertime.on("panstart", function (e) {
                let mousePos = {
                    x: e.center.x,
                    y: e.center.y
                };
                EDITOR.chooseWidget(mousePos);

                if (!canMove()) return;

                EDITOR.currentWidget.moveData.panX = e.deltaX;
                EDITOR.currentWidget.moveData.panY = e.deltaY;
            }
        );


        hammertime.on("pan", function (e) {
                if (!canMove() || EDITOR.currentWidget.cropper === undefined) return;

                EDITOR.currentWidget.cropper.move(e.deltaX - EDITOR.currentWidget.moveData.panX,
                    e.deltaY - EDITOR.currentWidget.moveData.panY);

                EDITOR.currentWidget.moveData.panX = e.deltaX;
                EDITOR.currentWidget.moveData.panY = e.deltaY;
            }
        );

        hammertime.on("panend", function (e) {
                if (!canMove()) return;

                EDITOR.currentWidget.moveData.panX = e.deltaX;
                EDITOR.currentWidget.moveData.panY = e.deltaY;

                EDITOR.doImageCrop(EDITOR.currentId, EDITOR.currentWidget);
            }
        );
    }


    /**
     * 根据鼠标触发坐标，选择命中组件
     *
     * 此函数基于 仿射变换矩阵，来判断鼠标坐标是否在控件上
     * */
    chooseWidget(mousePos) {
        let EDITOR          = this;
        let containerOffset = EDITOR.container.offset();
        let chosenWidget    = null;
        let hitIndex        = -1;
        let scrollTop       = $(document).scrollTop();
        let scrollLeft      = $(document).scrollLeft();

        mousePos = {
            x: mousePos.x + scrollLeft,
            y: mousePos.y + scrollTop
        };

        for (let widget of EDITOR.groupWidgets[EDITOR.currentGroupId]) {

            let asset = widget.asset;

            // 图片组件选中有两种情况
            // 1. 没有初始化截图组件 使用 widget.asset 中的位置尺寸信息计算
            // 2. 已经初始化截图组件 使用 widget.cropperBoxNode 计算
            if (widget.cropper === undefined) {

                if (widget.type === 2) { // 文字组件
                    let pos = {
                        x: widget.asset.displayOffsetX + containerOffset.left,
                        y: widget.asset.displayOffsetY + containerOffset.top
                    };

                    let affine = new Affine();
                    affine.set({x: 0, y: 0}, pos, {x: 1, y: 1}, 0);

                    let inverse = affine.getInverse();
                    let pointer = inverse.transform(mousePos);
                    if (pointer.x > 0 && pointer.x < asset.originWidth * asset.displayScale
                        && pointer.y > 0 && pointer.y < asset.originHeight * asset.displayScale) {
                        if (asset.ui.index > hitIndex) {
                            hitIndex     = asset.ui.index;
                            chosenWidget = widget;
                        }
                    }
                } else { // 图片组件

                    let pre = new Affine();
                    pre.mE  = containerOffset.left;
                    pre.mF  = containerOffset.top;

                    let affine = new Affine();
                    affine.cloneFrom(widget.matrix);
                    affine.multiply(pre);

                    let inverse = affine.getInverse();
                    let pointer = inverse.transform(mousePos);
                    if (pointer.x > 0 && pointer.x < asset.originWidth
                        && pointer.y > 0 && pointer.y < asset.originHeight) {
                        if (asset.ui.index > hitIndex) {
                            hitIndex     = asset.ui.index;
                            chosenWidget = widget;
                        }
                    }
                }
            } else {
                let cropImageData = widget.cropper.getImageData();

                let pre = new Affine();
                pre.mE  = containerOffset.left;
                pre.mF  = containerOffset.top;

                let affine = new Affine();
                affine.cloneFrom(cropImageData.matrix);
                affine.multiply(pre);

                let inverse = affine.getInverse();
                let pointer = inverse.transform(mousePos);

                if (pointer.x > 0 && pointer.x < cropImageData.options.originWidth
                    && pointer.y > 0 && pointer.y < cropImageData.options.originHeight) {
                    if (asset.ui.index > hitIndex) {
                        hitIndex     = asset.ui.index;
                        chosenWidget = widget;
                    }
                }
            }
        }

        if (chosenWidget) {
            EDITOR.currentWidget = chosenWidget;
            EDITOR.currentId     = chosenWidget.id;

        } else {
            EDITOR.currentWidget = null;
            EDITOR.currentId     = null;
        }
    };


    /**
     * 获取需要准备好的素材数目
     *
     * */
    getShouldReadyAssetsNum() {
        return this.currentOrders.length;
    }


    /**
     * 获取已经准备好的素材数目
     * */
    getReadyAssetNum() {
        let count = 0;
        for (let i in this.editedAssets) {
            count++;
        }
        return count;
    }


    /**
     * 获取编辑器当前的数据
     * */
    getEditorData() {
        return {
            data: this.croppedData
        };
    }

    // 打印调试信息
    log(text) {
        if (this.enable_debug_box) {
            this.debug_count++;
            this.debug_box.prepend("<p>" + this.debug_count + " : " + text + "</p>");
        }
        if (this.debug) {
            console.log(text);
        }
    }

    // 解析 template json
    parseConfigJson(json) {
        for (let asset of json.assets) {
            let key = asset.key;
            if (asset.ui) {
                let ui = asset.ui;

                asset.editable        = true; // 是否可编辑
                asset.moveable        = true; // 是否可移动
                asset.default_text    = "";   //  默认文字
                asset.default_img_src = "";   // 默认图像链接

                // 特殊设置  素材属性
                if (this.assetSettings && this.assetSettings[key]) {
                    if (typeof (this.assetSettings[key].editable) !== "undefined") {
                        asset.editable = this.assetSettings[key].editable;
                    }
                    if (typeof (this.assetSettings[key].moveable) !== "undefined") {
                        asset.moveable = this.assetSettings[key].moveable;
                    }
                    if (typeof (this.assetSettings[key].text) !== "undefined") {
                        asset.default_text = this.assetSettings[key].text;
                    }
                    if (typeof (this.assetSettings[key].img_src) !== "undefined") {
                        asset.default_img_src = this.assetSettings[key].img_src;
                    }
                }


                this.currentAssets[asset.key] = asset;
                //  1 - 图片， 2-文字
                if (typeof(this.currentGroups[ui.group]) === "undefined") {
                    this.currentGroups[ui.group] = [];
                }
                this.currentGroups[ui.group].push(asset.key);
                this.currentOrders.push(asset.key);
            }
        }

        if (this.debug) {
            console.log(this.currentGroups);
            console.log(this.currentOrders);
            console.log(this.currentAssets);
        }
    }


    // 获取组件　groups
    getGroups() {
        return this.currentGroups;
    }


    /**
     * 获取排序好的 url 字符串, 提交任务的数据对顺序有严格要求, 必须与 config json 中的顺序保持一致，没有则用空字符串替代
     *
     * @param uploaded_files 为对象，里面存储了 id => filekey 的映射关系
     * */
    getUrlsStr(uploaded_files) {
        let urls = "";
        for (let id of this.currentOrders) {
            if (uploaded_files[id]) {
                urls += urls === "" ? uploaded_files[id] : "," + uploaded_files[id];
            } else {
                urls += urls === "" ? "" : "," + "";
            }
        }
        return urls;
    }


    // 初始化 group
    initEditGroup(group_id) {
        if (!this.groupWidgets[group_id]) {
            this.groupWidgets[group_id] = [];

            let groupBox = $('<div class="group-wrapper" style="display: none"></div>');
            groupBox.attr("data-group-id", group_id);
            this.container.append(groupBox);

            let emptyNode = $('<div style="display:none;" ></div>');
            groupBox.append(emptyNode);

            this.groupBoxes[group_id] = {
                groupBox: groupBox,
                emptyNode: emptyNode
            };

            for (let key of this.currentGroups[group_id]) {
                let asset = this.currentAssets[key];
                if (asset.ui.type === 1) { // 图片
                    this.groupWidgets[group_id].push(this.createImageEditButton(key));
                } else { // 文字
                    this.groupWidgets[group_id].push(this.createTextEditButton(key));
                }
            }
            return true;
        }

        return false;
    }


    // 根据指定 group_id, 创建编辑页面
    createEditGroup(group_id) {
        // 隐藏其他widget
        this.container.find(".group-wrapper").hide();
        this.currentGroupId = group_id;

        // 初始化组编辑界面
        this.initEditGroup(group_id);

        // 根据素材尺寸调整编辑器大小
        // 1. 获取当前素材的尺寸
        // 2. 设置当前 editor 的宽高
        let firstChild = this.currentGroups[group_id][0];
        let setting    = this.getEditorSetting(this.currentAssets[firstChild]);
        if (this.options.resizeAnimate) {
            this.container.animate({width: setting.width + "px", height: setting.height + "px"});
        } else {
            this.container.css({width: setting.width + "px", height: setting.height + "px"});
        }
        this.groupBoxes[this.currentGroupId].groupBox.show();
    }


    // 创建文字编辑控件
    createTextEditButton(id) {

        let EDITOR = this;
        if (EDITOR.widgets[id]) {
            EDITOR.widgets[id].node.show();
            return;
        } else {
            EDITOR.widgets[id] = true;
        }


        // 计算截图框大小
        let asset        = this.currentAssets[id];
        let ui           = asset.ui;
        let originWidth  = asset.size[0];
        let originHeight = asset.size[1];

        let setting = this.getEditorSetting(asset);
        let scale   = setting.scale;

        // 容器宽高
        let editorWidth  = setting.width;
        let editorHeight = setting.height;

        // 素材编辑框
        let displayWidth  = ui.area[2] * scale;
        let displayHeight = ui.area[3] * scale;

        // calculate position
        let offsetX = ui.area[0];
        let offsetY = ui.area[1];

        let displayOffsetX = offsetX * scale;
        let displayOffsetY = offsetY * scale;

        asset.editorWidth  = editorWidth;
        asset.editorHeight = editorHeight;

        // store calc result
        asset.displayWidth  = displayWidth;
        asset.displayHieght = displayHeight;

        asset.offsetX = offsetX;
        asset.offsetY = offsetY;

        asset.displayOffsetX = displayOffsetX;
        asset.displayOffsetY = displayOffsetY;

        asset.originWidth  = ui.area[2];
        asset.originHeight = ui.area[3];

        asset.displayScale = scale;

        /**
         "ui_key":"", \
         "group": 1, \
         "index": 1, \
         "duration": 226, \
         "type": 2, \
         "default": "文字内容", \
         "max": 4, \
         "font": "Microsoft YaHei", \
         "size": 60, \
         "fill": "#f09fff", \
         "stroke": "#000000", \
         "width": 0, \
         "align": 0, \
         "shadow": "#000000", \
         "s_alpha": 0.0, \
         "s_angle": 0, \
         "s_dist": 0, \
         "s_size": 0, \
         "area": [144, 826, 240, 59] \
         * */

        let style = "width: " + displayWidth + "px;"
            + "height: " + displayHeight + "px;"
            + "left: " + displayOffsetX + "px;"
            + "top: " + displayOffsetY + "px;"
            + "z-index:" + asset.ui.index;


        let textAligns = {
            0: "left",
            1: "right",
            2: "center"
        };

        let input_style = "font-family:\"" + ui.font + "\""
            + ";font-size:" + ui.size * scale + "px"
            + ";color:" + ui.fill
            + ";-webkit-text-stroke:" + ui.width * scale + "px " + ui.stroke
            + "; text-align:" + textAligns[ui.align]
        ;

        let option = {
            style: style,
            input_style: input_style,
            input_length: ui.max,
            input_default: asset.default_text.length === 0 ? ui.default : asset.default_text,
        };

        let widget      = {};
        widget.id       = id;
        widget.moveable = false; // 是否可移动
        widget.editable = asset.editable; // 是否可编辑
        widget.asset    = this.currentAssets[id];
        widget.group_id = widget.asset.ui.group;
        widget.type     = widget.asset.ui.type;

        let $text_box    = $(this.htmls.text_box_html);
        widget.node      = $text_box.attr("style", option.style).attr("widget-id", id);
        widget.inputNode = $text_box.find(".ve-text-input");
        widget.maskNode  = $(this.htmls.mask_html);

        widget.inputNode.attr("style", option.input_style)
            .attr("maxlength", option.input_length).val(option.input_default);

        if (widget.editable === false) {
            widget.inputNode.attr("readonly", "readonly");
        }


        // 设置遮罩层
        let mask_img_src = "";
        if (asset.ui.f && asset.ui.f.length > 1) {
            mask_img_src = asset.ui.f;
        }

        // 配饰图
        if (mask_img_src !== "") {
            let img    = new Image();
            img.src    = mask_img_src;
            img.onload = function (e) {
                widget.maskNode.css({
                    "background-image": "url(" + img.src + ") ",
                    "background-repeat": "no-repeat",
                    "z-index": widget.asset.ui.index,
                    "display": "block"
                }).show();
            };
        }


        widget.click = function (e) {
            e.stopPropagation();
        };

        widget.inputNode.on("focus", function (e) {
            if (widget.editable === false) return;
            if (widget.inputNode.val() === option.input_default) {
                widget.inputNode.val("");
            }
        });

        EDITOR.widgets[id] = widget;


        // 画布尺寸必须和原尺寸保持一致
        let canvasWidthInPx  = asset.size[0];// this.globalAssetRatio;
        let canvasHeightInPx = asset.size[1];//  this.globalAssetRatio;
        let canvasHtml       = '<canvas style="display: none" width="' + canvasWidthInPx + '" height="' + canvasHeightInPx + '"></canvas>';
        let $canvas          = $(canvasHtml);

        $("body").append($canvas);
        widget.canvas = $canvas;

        // 绘制文字并截图
        let doTextCropAction = function (e) {
            let text    = widget.inputNode.val();
            let options = {
                width: asset.size[0], // canvas width
                height: asset.size[1], // canvas height
                text: text,
                textAlign: textAligns[ui.align],
                fontColor: ui.fill,  // 文字颜色
                fontStroke: ui.stroke, // 文字描边颜色
                fontStrokeWidth: ui.width, // 文字描边宽度
                fontFamily: ui.font,
                fontSize: ui.size,
                lineHeight: ui.line_height ? ui.line_height : ui.size + 6, // 行高
                x: ui.area[0], // 文字绘制 x 轴位置，相对于 canvas 左上角
                y: ui.area[1], // 文字绘制 y 轴位置,
                textWidth: ui.area[2], // 文字框宽度
                textHeight: ui.area[3], // 文字框高度
            };

            // 绘制文字
            EDITOR.fillTextCanvas(widget.canvas.get(0), options);

            // 获取截图内容
            EDITOR.doTextCrop(id, widget, function () {
                EDITOR.editedAssets[id] = true;
            });
        };

        // 失去焦点事件
        widget.inputNode.on("blur", doTextCropAction);
        // 文本改变事件
        widget.inputNode.bind('input propertychange', doTextCropAction);

        // widget
        widget.doTextCropAction = doTextCropAction;

        widget.parentNode = this.container;

        widget.node.insertAfter(this.groupBoxes[widget.group_id].emptyNode);
        widget.maskNode.insertBefore(this.groupBoxes[widget.group_id].emptyNode);

        // ready
        doTextCropAction();

        return widget;
    }


    // 获取文字截图
    doTextCrop(id, widget, callback) {
        let EDITOR = this;
        EDITOR.log("do crop : " + id);
        let canvas = widget.canvas.get(0);

        if (canvas) {

            if (this.debug) {
                let img = new Image();
                img.src = canvas.toDataURL("image/png");
                $("body").append(img);
                img = null;
            }

            canvas.toBlob(function (blob) {
                if (!EDITOR.croppedData[id]) {
                    EDITOR.croppedCount++;
                }
                let del = delete EDITOR.croppedData[id];
                EDITOR.log("delete crop data : " + del);
                EDITOR.croppedData[id] = null;
                canvas                 = null;

                EDITOR.croppedData[id] = blob;

                if (callback) {
                    callback();
                }
            }, "image/png", 1);
        }
    }

    // 绘制文字到画布
    fillTextCanvas(canvas, options) {
        let context = canvas.getContext('2d');  //获取对应的2D对象(画笔)

        // 设置　canvas 宽度
        canvas.width  = options.width;
        canvas.height = options.height;

        // options.fontSize *= 0.8;
        // 设置字体
        context.fillStyle   = options.fontColor; // 文字颜色
        context.strokeStyle = options.fontStroke;  // 描边颜色
        context.font        = options.fontSize + "px " + options.fontFamily;


        let text      = options.text;
        let textWidth = context.measureText(text).width;
        let x         = options.x;
        let y         = options.y;

        // @TODO 多行字体对齐优化
        if (options.textAlign === "center") {
            if (options.width > textWidth) {
                x += ( options.textWidth - textWidth) / 2;
            }
        } else if (options.textAlign === "right") {
            if (options.width > textWidth) {
                x += options.textWidth - textWidth;
            }
        }

        context.textBaseline = "top"; // 设置文字基线类型

        // 描边参数
        context.strokeWidth = options.fontStrokeWidth;
        context.strokeStyle = options.fontStroke;  // 描边颜色

        if (options.fontStrokeWidth > 0) {
            context.strokeText(options.text, options.x, options.y);
        }


        // 特殊格式处理
        let brIndex = text.indexOf("{br}");

        if (brIndex > 0) {
            let row = text.split("{br}");

            let maxBlockNum = 0;
            for (let b = 0; b < row.length; b++) {
                let chars = row[b].split("{space}");
                if (chars.length === 0) {
                    continue;
                }
                maxBlockNum = chars.length > maxBlockNum ? chars.length : maxBlockNum;
            }
            if (maxBlockNum === 0) {
                return;
            }


            let charLen = options.textWidth / maxBlockNum;
            for (let b = 0; b < row.length; b++) {
                let chars = row[b].split("{space}");
                if (chars.length === 0) {
                    continue;
                }

                for (let j = 0; j < chars.length; j++) {
                    let charStr = chars[j];
                    context.fillText(charStr, x + charLen * j, y + b * options.lineHeight);
                    if (options.fontStrokeWidth > 0) {
                        context.strokeText(charStr, x + charLen * j, y + b * (options.lineHeight));
                    }
                }
            }
        } else if (textWidth > options.textWidth) { // 多行
            let chr  = text.split("");
            let temp = "";
            let row  = [];
            for (let a = 0; a < text.length; a++) {
                if (context.measureText(temp).width < options.textWidth && context.measureText(temp + (chr[a])).width <= options.textWidth) {
                    temp += chr[a];
                }
                else {
                    row.push(temp);
                    temp = chr[a];
                }
            }
            row.push(temp);

            // context.fillText(row[0], 0, 0);
            // context.fillText(row[1], 0, options.fontSize);
            for (let b = 0; b < row.length; b++) {
                context.fillText(row[b], x, y + b * options.lineHeight);
                if (options.fontStrokeWidth > 0) {
                    context.strokeText(row[b], x, y + b * options.lineHeight);
                }
            }
        } else {
            context.fillText(text, x, y);
            if (options.fontStrokeWidth > 0) {
                context.strokeText(text, x, y);
            }
        }
    }

    getEditorSetting(asset) {

        let width = asset.size[0] < this.options.maxEditorWidth ? asset.size[0] : this.options.maxEditorWidth;
        let scale = width / asset.size[0];

        // 容器宽高
        let editorWidth  = width;
        let editorHeight = scale * asset.size[1];

        return {width: editorWidth, height: editorHeight, scale: scale};
    }

    // 创建图片编辑控件
    createImageEditButton(id) {
        let EDITOR = this;
        if (EDITOR.widgets[id]) {
            EDITOR.widgets[id].node.show();
            return;
        } else {
            EDITOR.widgets[id] = true;
        }

        // 计算截图框大小
        let asset        = this.currentAssets[id];
        let ui           = asset.ui;
        let originWidth  = asset.size[0];
        let originHeight = asset.size[1];

        let setting = this.getEditorSetting(asset);
        let scale   = setting.scale;

        // 容器宽高
        let editorWidth  = setting.width;
        let editorHeight = setting.height;

        // 获取展示宽高
        let displayWidth  = editorWidth;
        let displayHeight = editorHeight;

        // calculate position
        let offsetX = asset.ui.p[0] - asset.ui.a[0];
        let offsetY = asset.ui.p[1] - asset.ui.a[1];

        let displayOffsetX = offsetX * scale;//* asset.ui.s[0];
        let displayOffsetY = offsetY * scale;// * asset.ui.s[0];


        asset.editorWidth  = editorWidth;
        asset.editorHeight = editorHeight;

        asset.originWidth  = originWidth;
        asset.originHeight = originHeight;

        asset.displayWidth  = displayWidth;
        asset.displayHieght = displayHeight;

        asset.offsetX = offsetX;
        asset.offsetY = offsetY;

        asset.displayOffsetX = displayOffsetX;
        asset.displayOffsetY = displayOffsetY;

        asset.originScale  = scale;
        asset.displayScale = asset.ui.s[0] * scale;

        let anchorStr = 0 + "px " + 0 + "px";

        // 缩放矩阵
        let preAffine = new Affine();
        preAffine.mA  = scale;
        preAffine.mD  = scale;


        // 原始矩阵
        let anchor   = {x: ui.a[0], y: ui.a[1]};
        let position = {x: ui.p[0], y: ui.p[1]};
        let zoom     = {x: ui.s[1], y: ui.s[1]};
        let rotation = ui.r;

        let origin = new Affine();
        origin.set(anchor, position, zoom, rotation);
        origin.multiply(preAffine);

        let matrix = origin.mA + ", " + origin.mB + "," + origin.mC + ", " + origin.mD + "," + origin.mE + "," + origin.mF;
        let style  = "width: " + asset.size[0] + "px"
            + "; height: " + asset.size[1] + "px"
            + "; z-index:" + asset.ui.index
            + "; transform: matrix(" + matrix + ")"
            + "; -webkit-transform-origin:" + anchorStr;


        let option = {
            style: style,
            mask_img_src: ""
        };

        // 设置遮罩层
        if (asset.ui.f && asset.ui.f.length > 1) {
            option.mask_img_src = asset.ui.f;
        }

        let widget = this.buildImageEditButton(id, option);

        widget.matrix      = origin;
        widget.moveData    = {};
        EDITOR.widgets[id] = widget;
        widget.parentNode  = this.container;

        widget.node.insertBefore(this.groupBoxes[widget.group_id].emptyNode);
        widget.maskNode.insertAfter(this.groupBoxes[widget.group_id].emptyNode);
        return widget;
    }


    // 截图保存数据
    doImageCrop(id, widget, callback) {
        let EDITOR = this;
        EDITOR.log("do crop : " + id + " cropper : " + typeof (widget.cropper));

        if (widget.cropper === undefined) return;
        let cropper = widget.cropper;

        let asset  = EDITOR.currentAssets[id];
        let canvas = cropper.getCroppedCanvas({
            width: asset.size[0],
            height: asset.size[1]
        });

        if (canvas) {
            if (EDITOR.debug) {
                let img = new Image();
                img.src = canvas.toDataURL("image/png");
                $("body").append(img);
                img = null;
            }

            // 转成 blob
            canvas.toBlob(function (blob) {
                if (!EDITOR.croppedData[id]) {
                    EDITOR.croppedCount++;
                }
                delete EDITOR.croppedData[id];

                EDITOR.croppedData[id] = null;
                canvas                 = null;

                EDITOR.croppedData[id] = blob;
                EDITOR.log(EDITOR.croppedData);

                EDITOR.editedAssets[id] = true;

                if (callback) {
                    callback();
                }
            }, "image/png", 1);
        }
    }

    createImageCropper(id, widget) {
        let EDITOR = this;
        let asset  = this.currentAssets[id];

        let cropperOptions = {
            containerWidth: 200,
            containerHeight: 400,
            cropperWidth: 200,
            cropperHeight: 400,
            originWidth: asset.size[0],
            originHeight: asset.size[1],
            zIndex: asset.ui.index,
            asset: asset,
            matrix: widget.matrix,
            originScale: asset.originScale
        };

        cropperOptions.containerWidth  = EDITOR.options.containerWidth;
        cropperOptions.containerHeight = EDITOR.options.containerHeight;
        cropperOptions.cropperWidth    = asset.originWidth;
        cropperOptions.cropperHeight   = asset.originHeight;

        if (widget.cropper) {
            widget.cropper.destroy();
            widget.cropper = null;
        }

        cropperOptions.ready = function (cropper) {
            cropper.cropBox.insertBefore(EDITOR.groupBoxes[EDITOR.currentGroupId].emptyNode);
            widget.cropperBoxNode = cropper.cropBox;

            EDITOR.doImageCrop(id, widget);
        };

        let image_to_crop = widget.cropImgNode.get(0);
        widget.cropper    = new VCropper(image_to_crop, cropperOptions);
        widget.moveData   = widget.cropper.moveData;
        widget.node.hide();
    }


    // 构建 图片编辑控件
    buildImageEditButton(id, option) {
        let EDITOR      = this;
        let widget      = {};
        widget.id       = id;
        widget.asset    = this.currentAssets[id];
        widget.moveable = widget.asset.moveable;
        widget.editable = widget.asset.editable;
        widget.group_id = widget.asset.ui.group;
        widget.type     = widget.asset.ui.type;

        let $img_box         = $(this.htmls.img_box_html);
        widget.node          = $img_box.attr("style", option.style).attr("widget-id", id);
        widget.imgButtonNode = $img_box.find(".ve-img-button").attr("id", id);//.attr("style", option.style);
        widget.inputNode     = $img_box.find("input");
        widget.maskNode      = $(this.htmls.mask_html);


        // @TODO release
        widget.destroy = function () {
            // release cropper
            // release hammer
        };

        // 图片控件点击事件
        widget.click = function (e) {
            e.stopPropagation();
            if (!widget.editable) {
                return;
            }
            widget.inputNode.trigger('click');
        };

        widget.inputNode.on("click", function (e) {
            e.stopPropagation();
        });

        // 配饰图
        if (option.mask_img_src !== "") {
            let img    = new Image();
            img.src    = option.mask_img_src;
            img.onload = function () {
                widget.maskNode.css({
                    "background-image": "url(" + img.src + ") ",
                    "background-repeat": "no-repeat",
                    "z-index": widget.asset.ui.index,
                    "display": "block"
                }).show();
            };
        }

        // 预设图片
        if (widget.asset.default_img_src.length > 0) {
            let img = new Image();
            img.url = widget.asset.default_img_src;

            let $img = $(img).attr("src", img.url);
            $(img).css({display: "none"});
            $("body").append($img);

            widget.cropImgNode = $img;
            // 构建截图器
            img.onload         = function (e) {
                EDITOR.createImageCropper(id, widget);
            };
        }


        // 是否可编辑
        if (widget.editable) {
            widget.inputNode.on("change", function (e) {

                let target = e.target;
                let done   = function (url) {
                    let $crop_box      = $(EDITOR.htmls.crop_html);
                    widget.cropImgNode = $($crop_box.find(".ve-img-to-crop").get(0));
                    widget.cropImgNode.attr("src", url);

                    // 删除旧图
                    if (widget.imgContainerNode) {
                        widget.imgContainerNode.remove();
                    }

                    widget.imgContainerNode = $crop_box;
                    widget.imgButtonNode.append($crop_box);

                    // 构建截图器
                    EDITOR.createImageCropper(id, widget);
                };


                // load local file
                let files = target.files;
                let reader;
                let file;
                if (files && files.length > 0) {
                    file = files[0];
                    if (URL) {
                        done(URL.createObjectURL(file));
                    } else if (FileReader) {
                        reader = new FileReader();
                        reader.readAsDataURL(file);
                        reader.onload = function () {
                            done(reader.result);
                        };
                    }
                }
            });

        }
        return widget;
    }


    // 初始化 html
    initHtml() {
        let html          = {};
        html.img_box_html = `<div class="ve-img-box ve-widget">
            <div class="ve-img-button" ></div>
            <input type="file" class="ve-select-image" name="image" accept="image/*" style="display: none" />
            </div>`;
        html.mask_html = `<div class="ve-widget-mask" style="display:none"></div>`;
        html.crop_html = `<div class="ve-img-container"><img class="ve-img-to-crop" ></div>`;
        html.debug_box_html = '<div class="ve-debug-box" ><span class="ve-empty-line"></span></div>';
        html.text_box_html =
            `<div class="ve-text-box ve-widget"><div class="ve-text-container">
                <textarea class="ve-text-input" name="ve-text" > </textarea>
            </div></div>`;
        this.htmls = html;
    }

}
