# VE 视频引擎，h5 sdk/demo

该程序提供了一个简易的编辑器，在用户编辑上传好素材后，可以提交任务到 VECloud 提供的渲染服务，
等待渲染完成后，用户可以看到渲染的视频。

**测试程序使用了 VECloud 的存储和测试渲染服务，VECloud 会定期清空测试数据和文件**

### 使用步骤
+ 打开 [首页](www.seeshiontech.com) 注册用户，登录后在 "控制台->账号信息" 获取用户ID
+ 联系网站底部的客服人员获取渲染资格和余额
+ 打开 ```index.html``` 找到 veUid, 替换成自己的用户 ID，即可开启测试

### 标准模板基本步骤

1. 读取模板 config.json 文件，解析后获取 assets 数组
2. 遍历 assets 中有 ui 字段的元素
3. 根据 group 字段分组，每个 group 是一个编辑界面
4. 遍历 group 中的元素，根据元素类型，位置，缩放，移动 生成编辑界面
5. 按照给定的配置摆放图片素材，并截图
6. 按照给定的配置摆放文字，并截图
7. 获取编辑好的素材数据提交到服务器

### 2.0适配模板
1. 模板 config.json 中 ```ui_version: 2.x```, 不适用于**1.0版本**

### 动态模板基本步骤

1. 初始化路由(```initRoute()```), 为了循环出所需的模板列表, 并添加点击事件
2. 点击图片上传事件(```handleMultipleUpload()```), 处理数据(```handleData()```)
3. 处理数据之后, 初始化七牛token(```initQiNiuToken()```), 初始化input输入框(```initInput()```), 初始化导航栏(```initNavigation()```)
4. 提交数据(```onSubmit()```), 同步等待七牛异步上传成功回调(```qiNiuUploadImage()```)
5. 轮询接口(```socketInterFace()```)
6. 播放视频

### API
```
配置信息对象
let options = {
        container_id: "ve_editor",             // 必传容器id
        containerWidth: width,                // 容器宽度
        containerHeight: height,              // 容器高度
        maxContainerWidth: width,             // 编辑器最大宽度
        debug: false,                         // 开启调试截图组件
        enable_debug_box: false,              // 调试模式
        config_json: tpl_config_json,         // 模板配置文件路径
        TemplateUrl: TemplateUrl,             // 模板路径
        backgroundOpacity: backgroundOpacity  // 背景图透明度
}
let veditor = new VEditor(options);
```


+ 属性列表
    | 属性名 | 备注                                                  |
    | --------------------- |-------------------------------------  |
    | veditor.currentGroups | `解析模板得到的分组对象`                |
    | veditor.currentGroups | `解析模板得到的分组对象`                |
    | veditor.globalArray   | `全局带ui的数组`                       |
    | veditor.widgets       | `已经创建的组件`                       |
    | veditor.currentAssets | `当前素材配置`                         |
    | veditor.currentOrders | `素材顺序`                             |
    | veditor.currentGroups | `当前 groups`                         |
    | veditor.groupWidgets  | `存储组 widgets`                      |
    | veditor.groupBoxes    | `存储组 wrapper box`                  |
    | veditor.croppedData   | `完成截图的数据`                       | 
    | veditor.croppedCount  | `截图计数`                             |
    | veditor.editedAssets  | `素材是否编辑标记 map, 已经编辑过的素材将会被设置为 true` |
    | veditor.uiGroup       | `存储全局group size`                   |
    | veditor.dedup         | `存储字符动画数组`                      |

+ 方法列表
    | 方法名                               | 参数          | 返回值              | 备注                                                                                                                       |
    | ----------------------------------- | ------------- | ------------------- | ------------------------------------------------------------------------------------------------------------------------- |
    | createEditGroup(group_id)   | Number 分组ID       | 无                        | 根据指定 group_id, 创建编辑页面                                                                               |
    | appendBackground()          | 无                  | 无                        | 插入背景图到页面                                                                                            |
    | getReadyAssetNum()          | 无                  | Number 准备好的素材数目    | 获取已经准备好的素材数目                                                                                      |
    | getEditorData()             | 无                  | Object 编辑器当前的数据    | 获取编辑器当前的数据                                                                                          |
    | log()                       | 无                  | 无                        | 打印调试信息                                                                                                 |
    | getUrlsStr(uploaded_files)  | Object 素材对象      | String 排序好的 url 字符串 | 获取排序好的 url 字符串, 提交任务的数据对顺序有严格要求, 必须与 config json 中的顺序保持一致，没有则用空字符串替代 |
    | getTemplateFonts(configJson)| Object 模板配置文件  | Array 去重后的字体数组      | 获取模板用到的字体                                                                                          |
    | getEditorInfo()             | 无                  | Object 缩放宽高, 缩放比例   | 根据素材获取缩放宽高, 缩放比例                                                                               |
    | parseConfigJson(configJson) | Object 模板配置文件  | 无                         | 解析 template json                                                                                        |

### 前景图（背景图）路径说明

1. http://ve-oss1.atvideo.cc/VE_TEST__/ui/ui0.png (路径示例)
2. 前景图或背景图位于模板文件的ui目录中，添加前景图（背景图），需要使用模板路径 （TemplateUrl）+ ui/ + ui.b(ui.f)。





### 开发文档
- [标准模板 ui2.0详解](http://www.seeshiontech.com/docs/page_107.html)
- [模板素材替换规范](http://www.seeshiontech.com/docs/page_96.html)

### 项目结构
    
    |- docs 开发文档
    |- index.html   h5 导航页
    |- static.html  标准模板demo
    |- dynamic.html 动态模板demo
    |- lib/   相关依赖库
        |- vuploader.js 基于七牛的图片上传组件
        |- veditor.js 编辑器主文件
        |- vropper.js 图片截图组件
        |- affine.js 仿射变化矩阵计算库
        |- hammer.min.js 手势依赖库
        
### 运行示例
        
使用浏览器(推荐Chrome)打开, 并调出开发工具模拟手机,按步骤操作,即可看到渲染结果
