/**
 * Created by slayer on 18-12-28.
 *
 * @author dongyado
 */
class VCropper {

    constructor(imageNode, options, vm) {
        this.options = options;
        this.vm = vm;
        this.originImage = $(imageNode);
        this.originImageSrc = this.originImage.attr("src");
        this.debugBox = $("#debug_box");

        let cropper_html = '\
                <div class="ve-crop-img-box" style="display: none;"> \
                <img class="ve-crop-img" src="" style="display: none; width: 100%"/> \
                </div>';

        this.cropBox = $(cropper_html);
        this.cropImg = this.cropBox.find(".ve-crop-img");
        this.cropBox.css({
            width: options.cropperWidth + "px",
            height: options.cropperHeight + "px",
            zIndex: options.zIndex
        });

        // 设置变换矩阵
        this.matrix = new Affine();
        this.matrix.cloneFrom(this.options.matrix);
        this.canvasMatrix = new Affine();
        this.canvasMatrix.cloneFrom(this.matrix);


        // 图片参数
        this.cropImgOptions = {
            naturalWidth: 0,
            naturalHeight: 0,
            width: options.cropperWidth,
            height: options.cropperHeight,
            originWidth: options.cropperWidth,
            originHeight: options.cropperHeight,
            zIndex: options.zIndex,
            originScale: options.originScale
        };


        // 移动数据
        this.moveData = {
            rotation: 0,
            zoomScale: 0,
            panX: 0,
            panY: 0
        };


        this.originImage.addClass("ve-cropper-hidden");
        this.originImage.parent().css({ position: "relative" });
        this.cropImg.attr("src", this.originImageSrc);
        this.canvas = this.createCanvas();


        let CROPPER = this;
        this.isReady = false;
        this.orientation = 0;

        // 加载图片
        let img = new Image();
        if (this.originImageSrc.indexOf("http") === 0) { //
            img.setAttribute('crossOrigin', 'anonymous');
        }

        img.src = this.originImageSrc;
        img.onload = function () {
            CROPPER.loadedImg = img;
            CROPPER.isReady = true;
            CROPPER.cropImgOptions.naturalHeight = img.height;
            CROPPER.cropImgOptions.naturalWidth = img.width;

            let originWidth = CROPPER.cropImgOptions.width;
            let originHeight = (originWidth / img.width) * img.height;

            CROPPER.cropImgOptions.originWidth = originWidth;
            CROPPER.cropImgOptions.originHeight = originHeight;

            const { asset } = CROPPER.options;


            /**
             * ios 由于 orientation 的问题会导致 图片虽然是看着竖着，实际画出来是横的，所以需要预先旋转 90 度
             * */
            if (navigator.userAgent.indexOf('iPhone') > -1) {

                EXIF.getData(img, function () {

                    CROPPER.orientation = EXIF.getTag(this, 'Orientation');
                    if (CROPPER.orientation === 6) { // 顺时针 90

                        // iOS 需要按比例调换宽高，以修复选中区域不准确的问题
                        CROPPER.cropImgOptions.originWidth = originHeight * (originWidth / originHeight);
                        CROPPER.cropImgOptions.originHeight = originWidth * (originWidth / originHeight);

                        // 缩放
                        let zoomAffine = new Affine();
                        let zoomAffine2 = new Affine();
                        zoomAffine.set({ x: 0, y: 0 }, { x: 0, y: 0 }, {
                            x: originWidth / originHeight,
                            y: originWidth / originHeight
                        }, 0);
                        zoomAffine2.cloneFrom(zoomAffine);

                        CROPPER.canvasMatrix.multiply(asset.inverse); // 绘制乘以 逆矩阵

                        // CROPPER.matrix.premultiply(zoomAffine);
                        CROPPER.canvasMatrix.premultiply(zoomAffine2);

                        // 绕原点旋转90, 向右移动 originHeight
                        let rotateAffine = new Affine();
                        let rotateAffine2 = new Affine();
                        rotateAffine.set({ x: 0, y: 0 }, { x: originHeight, y: 0 }, { x: 1, y: 1 }, 90);
                        rotateAffine2.cloneFrom(rotateAffine);

                        // CROPPER.matrix.premultiply(rotateAffine);
                        CROPPER.canvasMatrix.premultiply(rotateAffine2);
                    }

                    CROPPER.ready();
                });
            } else {
                CROPPER.ready();
            }
        };

        if (this.options.bind) {
            this.bindListeners();
        }
    }


    /**
     * 销毁当前 cropper 数据和 element
     *
     * */
    destroy() {
        this.originImage.remove();
        this.cropBox.remove();
    }


    /**
     * 获取当前截图数据
     *
     * */
    getImageData() {
        return {
            options: this.cropImgOptions,
            matrix: this.matrix
        };
    }

    /**
     * 初始化完成回调
     *
     * */
    ready() {
        if (this.isReady === false) {
            return;
        }

        this.setTransform();
        this.cropBox.show();
        this.cropImg.show();

        // callback user custom ready function
        if (this.options.ready) {
            this.options.ready(this);
        }
    }


    /**
     * 获取 canvas
     *
     * */
    getCroppedCanvas(canvasOptions) {
        let CROPPER = this;
        this.canvas.width = canvasOptions.width;
        this.canvas.height = canvasOptions.height;
        let ctx = this.canvas.getContext("2d");

        let imgWidth = this.cropImgOptions.width;
        let imgHeight = this.cropImgOptions.naturalHeight / this.cropImgOptions.naturalWidth * this.cropImgOptions.width;

        let _zoom = canvasOptions.width / this.cropImgOptions.width;
        if (CROPPER.orientation === 6) {
            _zoom = canvasOptions.width / this.cropImgOptions.width;
        }
        let _scale = this.cropImgOptions.originScale;

        let anchor = { x: 0, y: 0 };
        let position = { x: 0, y: 0 };
        let scale = { x: 1 / _scale * _zoom, y: 1 / _scale * _zoom };
        let rotation = 0;
        let pre = new Affine();
        pre.set(anchor, position, scale, rotation);

        let origin2 = new Affine();
        origin2.cloneFrom(this.canvasMatrix);
        origin2.multiply(pre);
        origin2.multiply(canvasOptions.currentItem.inverse); // 绘制乘以 逆矩阵

        ctx.setTransform(origin2.mA, origin2.mB, origin2.mC, origin2.mD, origin2.mE, origin2.mF);
        ctx.drawImage(CROPPER.loadedImg, 0, 0, imgWidth, imgHeight);

        return this.canvas;
    }


    /**
     * 记录日志到 debug box
     * */
    log(text) {
        this.debugBox.append(text);
        this.debugBox.append("<br/>");
    }


    /**
     * 绑定编辑事件
     *
     * 默认不开启，一般选择在调用方重新实现
     *
     * */
    bindListeners() {
        let CROPPER = this;
        let hammertime = new Hammer(this.cropBox.get(0));
        hammertime.get('rotate').set({ enable: true });
        hammertime.get('pinch').set({ enable: true });

        hammertime.on("rotatestart", function (e) {
            CROPPER.moveData.rotation = e.rotation;
        });

        hammertime.on("rotateend", function (e) {
            CROPPER.moveData.rotation = 0;
        });

        hammertime.on("rotate", function (e) {
            let anchor = { x: e.center.x, y: e.center.y };
            CROPPER.rotate(e.rotation - CROPPER.moveData.rotation, anchor);
            CROPPER.moveData.rotation = e.rotation;
        });

        // 缩放
        hammertime.on("pinchstart", function (e) {
            CROPPER.moveData.zoomScale = e.scale;
        });

        hammertime.on("pinch", function (e) {
            let _zoom = e.scale / CROPPER.moveData.zoomScale;
            let anchor = { x: e.center.x, y: e.center.y };

            CROPPER.zoom(_zoom, anchor);
            CROPPER.moveData.zoomScale = e.scale;
        });

        // 移动
        hammertime.on("panstart", function (e) {
            CROPPER.moveData.panX = e.deltaX;
            CROPPER.moveData.panY = e.deltaY;
        });

        hammertime.on("pan", function (e) {
            CROPPER.move(e.deltaX - CROPPER.moveData.panX, e.deltaY - CROPPER.moveData.panY);
            CROPPER.moveData.panX = e.deltaX;
            CROPPER.moveData.panY = e.deltaY;
        });

        hammertime.on("panend", function (e) {
            CROPPER.moveData.panX = e.deltaX;
            CROPPER.moveData.panY = e.deltaY;
        });

    }


    // 缩放
    zoom(scale, anchorPointer) {
        const { asset } = this.options;
        let anchor = anchorPointer ? { x: anchorPointer.x - asset.canvasLeftOffset, y: anchorPointer.y - asset.canvasTopOffset } : { x: 0, y: 0 };
        let position = { x: anchor.x, y: anchor.y };
        let zoom = { x: scale, y: scale };
        let rotation = 0;

        let pre = new Affine();
        pre.set(anchor, position, zoom, rotation);
        this.matrix.multiply(pre);
        this.canvasMatrix.multiply(pre);

        this.setTransform();
        this.asyncSetData(this.vm);
    }


    // 移动
    move(x, y) {
        let anchor = { x: 0, y: 0 };
        let position = { x: x, y: y };
        let zoom = { x: 1, y: 1 };
        let rotation = 0;
        let pre = new Affine();
        pre.set(anchor, position, zoom, rotation);

        this.matrix.multiply(pre);
        this.canvasMatrix.multiply(pre);
        this.setTransform();
        this.asyncSetData(this.vm);
    }


    // 旋转
    rotate(rotation, anchorPointer) {
        const { asset } = this.options;
        let anchor = anchorPointer ? { x: anchorPointer.x - asset.canvasLeftOffset, y: anchorPointer.y - asset.canvasTopOffset } : { x: 0, y: 0 };
        let position = { x: anchor.x, y: anchor.y };
        let zoom = { x: 1, y: 1 };
        let pre = new Affine();

        pre.set(anchor, position, zoom, rotation);
        this.matrix.multiply(pre);
        this.canvasMatrix.multiply(pre);

        this.setTransform();
        this.asyncSetData(this.vm);
    }


    // 设置变换
    setTransform() {
        this.anchorStr = "0px 0px 0px";
        let origin = this.matrix;
        let matrix = origin.mA + ", " + origin.mB + "," + origin.mC + ", " + origin.mD + "," + origin.mE + "," + origin.mF;

        this.cropBox.css(
            {
                "width": this.cropImgOptions.width + "px",
                "height": this.cropImgOptions.height + "px",
                "z-index": this.cropImgOptions.zIndex,
                "transform": "matrix(" + matrix + ")",
                "-webkit-transform-origin": this.anchorStr
            }
        );
    }

    asyncSetData(vm) {
        var _temp = vm.backgroundArray.length ? { backgroundOpacity: 0.2 } : {};
        this.throttle(vm, { ..._temp })
    };

    throttle(vm, options) {
        var { backgroundOpacity } = { ...options };

        if (backgroundOpacity !== undefined) {
            var bg = {
                background: vm.backgroundArray !== undefined && vm.backgroundArray.length ? '' : '',
            };
            vm.container.css({ ...bg });
            document.querySelectorAll("#bga").forEach(function (item) {
                item.style.opacity = 0.2;
            });
        }

    };

    getStyle() {
        return this.anchorStr + "; " + this.transformStr;
    }

    createCanvas() {
        let img_canvas = document.createElement("canvas");
        $(img_canvas).insertAfter(this.cropBox).addClass("ve-cropper-canvas");
        return img_canvas;
    }

}
