/**
 * VEditor.
 *
 * @author dongyado
 */
class VEditor {

    constructor(options) {
        let EDITOR = this;
        this.options = {};

        // 校验数据
        if (!options.container_id) {
            throw new Error("container_id required");
        }

        this.container = $("#" + options.container_id);
        if (this.container.length === 0) {
            throw new Error("Invalid container_id : #" + container_id);
        }

        this.options.maxEditorWidth = options.containerWidth;
        this.options.containerWidth = !options.containerWidth ? this.container.width() : options.containerWidth;
        this.options.containerHeight = !options.containerHeight ? this.container.height() : options.containerHeight;

        this.container.html("");
        this.container.css({
            width: this.options.containerWidth + "px",
            height: this.options.containerHeight + "px",
            position: "relative",
            overflow: "hidden"
        });


        // 初始化常用的 html 变量
        this.initHtml();

        this.globalArray = []; // 全局带ui的数组
        this.widgets = {}; // 已经创建的组件
        this.currentAssets = {}; // 当前素材配置
        this.currentOrders = []; // 素材顺序
        this.currentGroups = {}; // 当前 groups
        this.groupWidgets = {}; // 存储组 widgets
        this.groupBoxes = {}; // 存储组 wrapper box
        this.croppedData = {}; // 完成截图的数据
        this.croppedCount = 0;  // 截图计数
        this.editedAssets = {}; // 素材是否编辑标记 map, 已经编辑过的素材将会被设置为 true
        this.uiGroup = []; // 存储全局group size
        this.charArray = []; // 存储字符动画数组
        this.TemplateUrl = options.TemplateUrl;
        this.bgOpacity = options.backgroundOpacity;

        // 调试配置
        this.debug = false;
        this.enable_debug_box = false;

        // append debug box
        if (options.enable_debug_box) {
            this.enable_debug_box = true;
            this.debug_box = $(this.htmls.debug_box_html);
            this.debug_count = 0;
            $("body").append(this.debug_box);
        }

        if (options.debug) {
            this.debug = options.debug;
        }

        // 初始化模板配置文件
        if (options.config_json) {
            this.uiGroup = options.config_json.ui_group;
            this.parseConfigJson(options.config_json);
        }

        // console.log(this.getTemplateFonts(options.config_json));
        // 当前组件和组件 id
        EDITOR.currentWidget = null;
        EDITOR.currentId = null;

        // 控件选择监听器
        let chooseWidgetListener = function (e) {
            let mousePos = {
                x: e.originalEvent.clientX,
                y: e.originalEvent.clientY
            };
            EDITOR.chooseWidget(mousePos);

            if (EDITOR.currentWidget) {
                EDITOR.currentWidget.click(e);
            }
        };

        // 为编辑器容器清除和重新绑定监听器
        this.container.unbind("click");
        this.container.bind("click", chooseWidgetListener);

        // 绑定事件
        this.bindEvents();
    }

    // 绑定控件编辑事件
    bindEvents() {
        let EDITOR = this;
        // 检查当前组件是否可用，是否可移动
        let canMove = function () {
            if (EDITOR.currentWidget === null
                || !EDITOR.currentWidget.moveable) {
                return false;
            }
            return true;
        };

        // 编辑事件绑定
        let hammertime = new Hammer(this.container.get(0));
        hammertime.get('rotate').set({ enable: true });
        hammertime.get('pinch').set({ enable: true });

        hammertime.get('pan').set({ direction: Hammer.DIRECTION_ALL });

        // 绑定旋转事件
        hammertime.on("rotatestart", function (e) {
            let mousePos = {
                x: e.center.x,
                y: e.center.y
            };
            EDITOR.chooseWidget(mousePos);

            if (!canMove()) return;

            EDITOR.currentWidget.moveData.rotation = e.rotation;
            EDITOR.log(EDITOR.currentWidget.cropper.getStyle());
        });

        hammertime.on("rotate", function (e) {
            if (!canMove() || EDITOR.currentWidget.cropper === undefined) return;
            if (e.rotation === 0) return;

            let deg = e.rotation - EDITOR.currentWidget.moveData.rotation;
            let anchor = { x: e.center.x, y: e.center.y };
            EDITOR.currentWidget.cropper.rotate(deg, anchor);
            EDITOR.currentWidget.moveData.rotation = e.rotation;
        }
        );

        hammertime.on("rotateend", function (e) {
            if (!canMove()) return;

            EDITOR.currentWidget.moveData.rotation = e.rotation;
            EDITOR.doImageCrop(EDITOR.currentId, EDITOR.currentWidget);
            EDITOR.log(EDITOR.currentWidget.cropper.getStyle());

            document.querySelectorAll("#bga").forEach((item, index, array) => {
                item.style.opacity = 1;
            });
        });

        // 缩放
        hammertime.on("pinchstart", function (e) {
            let mousePos = {
                x: e.center.x,
                y: e.center.y
            };
            EDITOR.chooseWidget(mousePos);

            if (!canMove()) return;

            EDITOR.currentWidget.moveData.zoomScale = e.scale;
        }
        );

        hammertime.on("pinch", function (e) {
            if (!canMove() || EDITOR.currentWidget.cropper === undefined) return;

            let zoom = e.scale / EDITOR.currentWidget.moveData.zoomScale;
            let anchor = { x: e.center.x, y: e.center.y };

            EDITOR.currentWidget.cropper.zoom(zoom, anchor);
            EDITOR.currentWidget.moveData.zoomScale = e.scale;
        }
        );

        hammertime.on("pinchend", function () {
            if (!canMove()) return;

            EDITOR.doImageCrop(EDITOR.currentId, EDITOR.currentWidget);

            document.querySelectorAll("#bga").forEach((item, index, array) => {
                item.style.opacity = 1;
            });
        }
        );

        // 移动
        hammertime.on("panstart", function (e) {
            let mousePos = {
                x: e.center.x,
                y: e.center.y
            };
            EDITOR.chooseWidget(mousePos);

            if (!canMove()) return;

            EDITOR.currentWidget.moveData.panX = e.deltaX;
            EDITOR.currentWidget.moveData.panY = e.deltaY;
        }
        );

        hammertime.on("pan", function (e) {
            if (!canMove() || EDITOR.currentWidget.cropper === undefined) return;

            EDITOR.currentWidget.cropper.move(e.deltaX - EDITOR.currentWidget.moveData.panX,
                e.deltaY - EDITOR.currentWidget.moveData.panY);

            EDITOR.currentWidget.moveData.panX = e.deltaX;
            EDITOR.currentWidget.moveData.panY = e.deltaY;
        }
        );

        hammertime.on("panend", function (e) {
            if (!canMove()) return;

            EDITOR.currentWidget.moveData.panX = e.deltaX;
            EDITOR.currentWidget.moveData.panY = e.deltaY;

            EDITOR.doImageCrop(EDITOR.currentId, EDITOR.currentWidget);

            document.querySelectorAll("#bga").forEach((item, index, array) => {
                item.style.opacity = 1;
            });
        }
        );
    }

    /**
     * 根据鼠标触发坐标，选择命中组件
     *
     * 此函数基于 仿射变换矩阵，来判断鼠标坐标是否在控件上
     * */
    chooseWidget(mousePos) {
        let EDITOR = this;
        let chosenWidget = null;
        let hitIndex = -1;
        let scrollTop = $(document).scrollTop();
        let scrollLeft = $(document).scrollLeft();

        mousePos = {
            x: mousePos.x + scrollLeft,
            y: mousePos.y + scrollTop
        };

        for (let widget of EDITOR.groupWidgets[EDITOR.currentGroupId]) {
            let asset = widget.asset;

            // 图片组件选中有两种情况
            // 1. 没有初始化截图组件 使用 widget.asset 中的位置尺寸信息计算
            // 2. 已经初始化截图组件 使用 widget.cropperBoxNode 计算
            if (widget.cropper === undefined) {

                if (widget.type === 2 || widget.type === 3) { // 文字组件

                    let affine = new Affine();

                    const { ui, originScale, editorWidth } = asset;

                    // 原始矩阵
                    let position = { x: ui.p[0], y: ui.p[1] };
                    let anchor = { x: ui.a[0], y: ui.a[1] };
                    let zoom = { x: ui.s[1], y: ui.s[1] };
                    let rotation = ui.r;

                    affine.set(anchor, position, zoom, rotation);

                    widget.matrix = affine;

                    widget.asset.inverse = affine.getInverse();
                    widget.asset.canvasLeftOffset = (window.screen.width - editorWidth) / 2;
                    widget.asset.canvasTopOffset = 20;
                    let pointer = mousePos;

                    let options = {
                        currentItem: asset,
                        mousePos,
                        offsetPos: {
                            x: (window.screen.width - editorWidth) / 2,
                            y: 80
                        }
                    };

                    if (EDITOR.handleCoordinate(options)) {
                        if (ui.index > hitIndex) {
                            hitIndex = ui.index;
                            chosenWidget = widget;
                        }
                    }
                } else { // 图片组件

                    // let affine = new Affine();
                    // affine.cloneFrom(widget.matrix);

                    const { ui, originScale, editorWidth } = asset;
                    // widget.asset.inverse = affine.getInverse();
                    widget.asset.canvasLeftOffset = (window.screen.width - editorWidth) / 2;
                    widget.asset.canvasTopOffset = 20;
                    let pointer = mousePos;

                    let options = {
                        currentItem: asset,
                        mousePos,
                        offsetPos: {
                            x: (window.screen.width - editorWidth) / 2,
                            y: 80
                        }
                    };

                    if (EDITOR.handleCoordinate(options)) {
                        if (ui.index > hitIndex) {
                            hitIndex = ui.index;
                            chosenWidget = widget;
                        }
                    }
                }
            } else {
                let cropImageData = widget.cropper.getImageData();

                let affine = new Affine();
                affine.cloneFrom(cropImageData.matrix);

                const { ui, originScale, editorWidth } = asset;
                // widget.asset.inverse = affine.getInverse();
                widget.asset.canvasLeftOffset = (window.screen.width - editorWidth) / 2;
                widget.asset.canvasTopOffset = 20;
                let pointer = mousePos;

                let options = {
                    currentItem: asset,
                    mousePos,
                    offsetPos: {
                        x: (window.screen.width - editorWidth) / 2,
                        y: 80
                    }
                };

                if (EDITOR.handleCoordinate(options)) {
                    if (ui.index > hitIndex) {
                        hitIndex = ui.index;
                        chosenWidget = widget;
                    }
                }
            }
        }

        if (chosenWidget) {
            EDITOR.currentWidget = chosenWidget;
            EDITOR.currentId = chosenWidget.id;
        } else {
            EDITOR.currentWidget = null;
            EDITOR.currentId = null;
        }
    };


    /**
     * 处理坐标
     *
     * @param options {object}
     * @return {boolean}
     * */
    handleCoordinate(options) {

        let { size } = this.uiGroup[Number(this.currentGroupId) - 1];
        let width = this.options.containerWidth;
        let scale = width / size[0];
        // 容器宽高
        let canvasW = width;
        let canvasH = scale * size[1];

        const { currentItem, mousePos, offsetPos } = options;
        let ui = currentItem.ui; // 选中当前item.ui
        let usw = canvasW / this.uiGroup[ui.group - 1].size[0];
        let ush = canvasH / this.uiGroup[ui.group - 1].size[1];

        // 缩放的四个坐标
        let scaleLeftTop = {
            x: ui.area[0] * usw,
            y: ui.area[1] * ush
        };
        let scaleRightTop = {
            x: ui.area[0] * usw + ui.area[2] * usw,
            y: ui.area[1] * ush
        };
        let scaleLeftBottom = {
            x: ui.area[0] * usw,
            y: ui.area[0] * ush + ui.area[3] * ush
        };
        let scaleRightBottom = {
            x: ui.area[0] * usw + ui.area[2] * usw,
            y: ui.area[1] * ush + ui.area[3] * ush
        };

        // 实际鼠标坐标
        let actualMouse = {
            x: mousePos.x - offsetPos.x,
            y: mousePos.y - offsetPos.y,
        };

        return actualMouse.x > scaleLeftTop.x &&
            scaleRightTop.x > actualMouse.x &&
            actualMouse.y > scaleLeftTop.y &&
            scaleRightBottom.y > actualMouse.y;

    };

    // 获取已经准备好的素材数目
    getReadyAssetNum() {
        let count = 0;
        for (let i in this.editedAssets) {
            count++;
        }
        return count;
    }

    // 获取编辑器当前的数据
    getEditorData() {
        return {
            data: this.croppedData
        };
    }

    // 打印调试信息
    log(text) {
        if (this.enable_debug_box) {
            this.debug_count++;
            this.debug_box.prepend("<p>" + this.debug_count + " : " + text + "</p>");
        }
        if (this.debug) {
            console.log(text);
        }
    }

    // 解析 template json
    parseConfigJson(json) {
        for (let asset of json.assets) {
            if (asset.ui) {
                let ui = asset.ui;
                this.globalArray.push(asset);
                asset.editable = true; // 是否可编辑
                asset.moveable = true; // 是否可移动

                this.currentAssets[asset.key] = asset;
                //  1 - 图片， 2-文字
                if (typeof (this.currentGroups[ui.group]) === "undefined") {
                    this.currentGroups[ui.group] = [];
                }
                this.currentGroups[ui.group].push(asset.key);
                this.currentOrders.push(asset.key);
            }
        }

        this.handleBackground(this.globalArray);

        if (this.debug) {
            console.log('currentGroups', this.currentGroups);
            console.log('currentOrders', this.currentOrders);
            console.log('currentAssets', this.currentAssets);
        }
    }


    /**
     * 解析前景图和背景图
     *
     * @param json {array}
     * @return {backgroundArray}
     * */
    async handleBackground(json) {
        let _tempB = [],      // 临时背景图数组
            _tempF = [],      // 临时前景图数组
            filterArray = [], // 过滤数组
            _tempArray = [],
            _this = this;

        // 过滤 带背景图, 前景图的数组
        json.forEach(x => {
            if (x.ui.f) {
                filterArray.push(x);
            }
            if (x.ui.b) {
                filterArray.push(x);
            }
        });

        json.map(async (data) => {
            const ui = data.ui;

            if (ui.f || ui.b) {
                if (!!ui.b) {

                    let options = {
                        data: data,
                        url: ui.b,
                        type: 'b',
                    };
                    _tempB = await this.ImageHandleResult(options);
                    _tempArray = [..._tempArray, ..._tempB]
                }

                if (!!ui.f) {

                    let options = {
                        data: data,
                        url: ui.f,
                        type: 'f',
                    };
                    _tempF = await this.ImageHandleResult(options);
                    _tempArray = [..._tempArray, ..._tempF]
                }
            }

            _tempArray = [..._tempArray];
        });

        let timer = null;
        clearInterval(timer);
        if (JSON.stringify(json) !== '{}') {
            timer = setInterval(() => {
                if (_tempArray.length === filterArray.length) {
                    this.backgroundArray = _tempArray;
                    _this.appendBackground();
                    clearInterval(timer);
                }
            }, 100)
        }

    };


    /**
     * 解析处理图片信息
     *
     * @param options {object}
     * @return Promise
     * */
    handleImageInfo(options) {
        const { uiGroup } = this;
        const { data, url, type } = options;
        const ui = data.ui;
        let _this = this;
        let _temp = [];
        let _tempB = [];
        let prefix = this.TemplateUrl;

        return new Promise(((resolve, reject) => {
            _this.getImageInfo(prefix + 'ui/' + url).then(res => {
                let width = this.options.containerWidth;
                let scale = width / uiGroup[ui.group - 1].size[0];
                // 容器宽高
                let canvasW = width;
                let canvasH = scale * uiGroup[ui.group - 1].size[1];

                let usw = canvasW / uiGroup[ui.group - 1].size[0];
                let ush = canvasH / uiGroup[ui.group - 1].size[1];
                _tempB.push({
                    calcWidth: res.width * usw,
                    calcHeight: res.height * ush,
                    width: res.width,
                    height: res.height,
                    b: prefix + 'ui/' + url,
                    z: type === 'b' ? 0 : 1,
                    o: 1,
                });

                _temp.push({
                    key: data.key,
                    name: data.name,
                    a: ui.a,
                    p: ui.p,
                    r: ui.r,
                    s: ui.s,
                    t: ui.t,
                    bf: _tempB,
                    type: ui.type,
                    area: ui.area,
                    editSize: ui.editSize,
                    b: prefix + 'ui/' + ui.b,
                    bZ: 0,
                    f: prefix + 'ui/' + ui.f,
                    fZ: 1,
                    group: ui.group,
                    index: ui.index,
                });
                resolve(_temp);
            }).catch(err => {
                console.log(err);
                reject(err);
            });
        }))
    };


    /**
     * 处理图片结果
     *
     * @param options {object}
     * @return Promise
     * */
    ImageHandleResult(options) {
        let _this = this;
        return new Promise(function (resolve, reject) {
            _this.handleImageInfo(options).then(res => {
                resolve(res);
            }).catch(err => {
                reject(err);
            });
        })
    };


    /**
     * 获取图片尺寸
     *
     * @param url {string}
     * @return Promise
     * */
    getImageInfo(url) {
        return new Promise((resolve => {
            let img = new Image();
            img.src = url;
            if (img.complete) {
                // 如果图片被缓存，则直接返回缓存数据
                resolve({
                    width: img.width,
                    height: img.height
                });
            } else {
                img.onload = function () {
                    resolve({
                        width: img.width,
                        height: img.height
                    });
                }
            }
        }))
    };


    /**
     * 插入背景图到页面
     *
     * */
    appendBackground() {
        const { backgroundArray, bgOpacity } = this;
        if (backgroundArray.length) {
            let bg = '';
            for (let j in backgroundArray) {
                if (Number(this.currentGroupId) - 1 === (backgroundArray[j].group - 1)) {

                    for (let k in backgroundArray[j].bf) {
                        bg += `
                    <img src=${backgroundArray[j].bf[k].b}
                           id='bga'
                           style='width: 100%;height: 100%;position: absolute;left: 0;top: 0;transition: all 0.3s;z-index: ${backgroundArray[j].bf[k].z};opacity: ${bgOpacity};${Number(this.currentGroupId) - 1 === (backgroundArray[j].group - 1) ? 'display:block' : 'display:none'}'
                    >
                    </img>
                    `
                    }
                }
            }
            this.groupBoxes[this.currentGroupId].groupBox.append(bg);
        }
    };

    /**
     * 获取排序好的 url 字符串, 提交任务的数据对顺序有严格要求, 必须与 config json 中的顺序保持一致，没有则用空字符串替代
     *
     * @param uploaded_files 为对象，里面存储了 id => fileKey 的映射关系
     * */
    getUrlsStr(uploaded_files) {
        let urls = "";
        for (let id of this.currentOrders) {
            if (uploaded_files[id]) {
                urls += urls === "" ? uploaded_files[id] : "," + uploaded_files[id];
            } else {
                urls += urls === "" ? "" : "," + "";
            }
        }
        return urls;
    }

    // 初始化 group
    initEditGroup(group_id) {
        if (!this.groupWidgets[group_id]) {
            this.groupWidgets[group_id] = [];

            let groupBox = $('<div class="group-wrapper" style="display: none"></div>');
            groupBox.attr("data-group-id", group_id);
            this.container.append(groupBox);

            let emptyNode = $('<div style="display:none;" ></div>');
            groupBox.append(emptyNode);

            this.groupBoxes[group_id] = {
                groupBox: groupBox,
                emptyNode: emptyNode
            };

            for (let key of this.currentGroups[group_id]) {
                let asset = this.currentAssets[key];
                if (asset.ui.type === 1) { // 图片
                    this.groupWidgets[group_id].push(this.createImageEditButton(key));
                } else { // 文字
                    this.groupWidgets[group_id].push(this.createTextEditButton(key));
                }
            }
            return true;
        }

        return false;
    }

    // 根据指定 group_id, 创建编辑页面
    createEditGroup(group_id) {
        // 隐藏其他widget
        this.container.find(".group-wrapper").hide();
        this.currentGroupId = group_id;

        // 初始化组编辑界面
        this.initEditGroup(group_id);

        // 根据素材尺寸调整编辑器大小
        // 1. 获取当前素材的尺寸
        // 2. 设置当前 editor 的宽高
        let firstChild = this.currentGroups[group_id][0];
        let setting = this.getEditorInfo(this.currentAssets[firstChild]);
        this.container.css({ width: setting.width + "px", height: setting.height + "px" });
        this.groupBoxes[this.currentGroupId].groupBox.show();
    }

    //获取模板用到的字体
    getTemplateFonts(configJson) {
        /**
        * 数组中对象去重
        * */
        let duplicate = (arr, type) => {
            let obj = {};
            let newArr = arr.reduce((cur, next) => {
                obj[next[type]] ? "" : obj[next[type]] = true && cur.push(next);
                return cur;
            }, []);
            return newArr;
        }

        let tempFonts = [];
        configJson.assets.forEach(x => {
            if (x.ui && x.ui.type !== 1) {
                tempFonts.push({
                    key: x.key,
                    family: x.ui.font_family,       //字体系列名称
                    font: x.ui.font,                //字体完整名称
                    fontStyle: x.ui.font_style,     //字体样式（例如粗体、细体或斜体等）
                    fontFile: x.ui.font_file,       //字体文件名称（不保证为所有字体类型返回字符串；对于某些类型的字体，返回值可能是空字符串）
                })
            }
        });
        return duplicate(tempFonts, 'family');
    }

    // 创建文字编辑控件
    createTextEditButton(id) {
        let EDITOR = this;
        if (EDITOR.widgets[id]) {
            EDITOR.widgets[id].node.show();
            return;
        } else {
            EDITOR.widgets[id] = true;
        }

        // 计算截图框大小
        let asset = this.currentAssets[id];
        let ui = asset.ui;
        // let setting = this.getEditorSetting(asset);
        let setting = this.getEditorInfo(asset);
        let scale = setting.scale;

        // 容器宽高
        let editorWidth = setting.width;
        let editorHeight = setting.height;

        // 素材编辑框
        let displayWidth = ui.area[2] * scale;
        let displayHeight = ui.area[3] * scale;

        let displayOffsetX = ui.area[0] * scale;
        let displayOffsetY = ui.area[1] * scale;

        asset.editorWidth = editorWidth;
        asset.editorHeight = editorHeight;

        // asset.originWidth  = ui.area[2];
        // asset.originHeight = ui.area[3];
        asset.originWidth = editorWidth;
        asset.originHeight = editorHeight;

        asset.displayScale = scale;

        let fIndex = asset.ui.index + 1; // FIXME

        let style = "width: " + displayWidth + "px;"
            + "height: " + displayHeight + "px;"
            + "left: " + displayOffsetX + "px;"
            + "top: " + displayOffsetY + "px;"
            + "z-index:" + fIndex;

        let textAligns = {
            0: "left",
            1: "right",
            2: "center"
        };

        let input_style = "font-family:\"" + ui.font + "\""
            + ";font-size:" + ui.size * scale + "px"
            + ";color:" + ui.fill
            + ";-webkit-text-stroke:" + ui.width * scale + "px " + ui.stroke
            + "; text-align:" + textAligns[ui.align]
            ;

        let option = {
            style: style,
            input_style: input_style,
            input_length: ui.max,
            input_default: ui.default,
        };
        let widget = {};
        widget.id = id;
        widget.moveable = false; // 是否可移动
        widget.editable = asset.editable; // 是否可编辑
        widget.asset = this.currentAssets[id];
        widget.group_id = widget.asset.ui.group;
        widget.type = widget.asset.ui.type;

        let $text_box = $(this.htmls.text_box_html);
        widget.node = $text_box.attr("style", option.style).attr("widget-id", id);
        widget.inputNode = $text_box.find(".ve-text-input");
        widget.maskNode = $(this.htmls.mask_html);
        widget.inputNode.attr("style", option.input_style)
            .attr("maxlength", option.input_length).val(option.input_default);

        if (widget.editable === false) {
            widget.inputNode.attr("readonly", "readonly");
        }

        // 设置遮罩层
        let mask_img_src = "";
        if (asset.ui.f && asset.ui.f.length > 1) {
            mask_img_src = asset.ui.f;
        }

        // 配饰图
        if (mask_img_src !== "") {
            let img = new Image();
            img.src = mask_img_src;
            img.onload = function () {
                widget.maskNode.css({
                    "background-image": "url(" + img.src + ") ",
                    "background-repeat": "no-repeat",
                    "z-index": widget.asset.ui.index,
                    "display": "block"
                }).show();
            };
        }

        widget.click = function (e) {
            e.stopPropagation();
        };

        widget.inputNode.on("focus", function () {
            if (widget.editable === false) return;
            if (widget.inputNode.val() === option.input_default) {
                // widget.inputNode.val("");
            }
        });

        EDITOR.widgets[id] = widget;

        // 画布尺寸必须和原尺寸保持一致
        let canvasWidthInPx = asset.size[0];// this.globalAssetRatio;
        let canvasHeightInPx = asset.size[1];//  this.globalAssetRatio;
        let canvasHtml = '<canvas style="display: none" width="' + canvasWidthInPx + '" height="' + canvasHeightInPx + '"></canvas>';
        let $canvas = $(canvasHtml);

        $("body").append($canvas);
        widget.canvas = $canvas;

        // 绘制文字并截图
        let doTextCropAction = function () {

            let text = widget.inputNode.val();
            let options = {
                width: asset.size[0],           // canvas width
                height: asset.size[1],          // canvas height
                text: text,
                textAlign: textAligns[ui.align],
                fontColor: ui.fill,             // 文字颜色
                fontStroke: ui.stroke,          // 文字描边颜色
                fontStrokeWidth: ui.width,      // 文字描边宽度
                fontFamily: ui.font,
                fontSize: ui.size,
                lineHeight: ui.line_height ? ui.line_height : ui.size + 6, // 行高
                x: ui.area[0],                  // 文字绘制 x 轴位置，相对于 canvas 左上角
                y: ui.area[1],                  // 文字绘制 y 轴位置,
                textWidth: ui.area[2],          // 文字框宽度
                textHeight: ui.area[3],         // 文字框高度
            };

            // 绘制文字
            EDITOR.fillTextCanvas(widget.canvas.get(0), options);

            // 获取截图内容
            EDITOR.doTextCrop(id, widget, function () {
                EDITOR.editedAssets[id] = true;
            });
        };

        // 字符动画
        let charArray = () => {
            let text = widget.inputNode.val();
            let chars = this.charArray;
            if (chars.length) {
                chars.map(x => {
                    if (x.id === widget.id && x.content !== text) {
                        x.content = text;
                    }
                });
                chars.push({
                    id: widget.id,
                    content: text,
                });
            } else {
                chars.push({
                    id: widget.id,
                    content: text,
                });
            }

            let obj = {};
            this.dedup = [];
            this.dedup = chars.reduce(function (item, next) {
                obj[next.id] ? '' : obj[next.id] = true && item.push(next);
                return item;
            }, []);
        };

        // 失去焦点事件
        widget.inputNode.on("blur", widget.type === 2 ? doTextCropAction : charArray);
        // 文本改变事件
        widget.inputNode.bind('input propertychange', widget.type === 2 ? doTextCropAction : '');
        // widget
        widget.doTextCropAction = widget.type === 2 ? doTextCropAction : '';

        widget.parentNode = this.container;
        //纵向文本
        if (ui.hasOwnProperty("box_text") && ui.editSize[1] > ui.editSize[0] && ui.direction === 1) {
            $(widget.node).css({
                "writing-mode": "vertical-rl"
            });
            $(widget.inputNode).css({
                "display": "table-cell"
            });
        }
        widget.node.insertAfter(this.groupBoxes[widget.group_id].emptyNode);
        widget.maskNode.insertBefore(this.groupBoxes[widget.group_id].emptyNode);


        if (widget.type === 2) {
            // ready
            doTextCropAction();
        }

        return widget;
    }

    // 2.0 版本 FIXME
    getEditorInfo() {
        let { size } = this.uiGroup[Number(this.currentGroupId) - 1];
        let width = this.options.containerWidth;
        let scale = width / size[0];
        // 容器宽高
        let editorWidth = width;
        let editorHeight = scale * size[1];

        return { width: editorWidth, height: editorHeight, scale: scale };
    }

    // 获取文字截图
    doTextCrop(id, widget, callback) {
        let EDITOR = this;
        EDITOR.log("do crop : " + id);
        let canvas = widget.canvas.get(0);

        if (canvas) {
            if (this.debug) {
                let img = new Image();
                img.src = canvas.toDataURL("image/png");
                $("body").append(img);
                img = null;
            }
            canvas.toBlob(function (blob) {
                if (!EDITOR.croppedData[id]) {
                    EDITOR.croppedCount++;
                }
                let del = delete EDITOR.croppedData[id];
                EDITOR.log("delete crop data : " + del);
                EDITOR.croppedData[id] = null;
                canvas = null;

                EDITOR.croppedData[id] = blob;

                if (callback) {
                    callback();
                }
            }, "image/png", 1);
        }
    }

    // 绘制文字到画布
    fillTextCanvas(canvas, options) {
        let context = canvas.getContext('2d');  //获取对应的2D对象(画笔)
        // 设置　canvas 宽度
        canvas.width = options.width;
        canvas.height = options.height;
        // 设置字体
        context.fillStyle = options.fontColor; // 文字颜色
        context.strokeStyle = options.fontStroke;  // 描边颜色
        context.font = options.fontSize + "px " + options.fontFamily;
        let text = options.text;
        let textWidth = context.measureText(text).width;
        let x = options.x;
        let y = options.y;
        // @TODO 多行字体对齐优化
        if (options.textAlign === "center") {
            if (options.width > textWidth) {
                x += (options.textWidth - textWidth) / 2;
            }
        } else if (options.textAlign === "right") {
            if (options.width > textWidth) {
                x += options.textWidth - textWidth;
            }
        }
        context.textBaseline = "top"; // 设置文字基线类型
        // 描边参数
        context.strokeWidth = options.fontStrokeWidth;
        context.strokeStyle = options.fontStroke;  // 描边颜色

        if (options.fontStrokeWidth > 0) {
            context.strokeText(options.text, options.x, options.y);
        }


        // 特殊格式处理
        let brIndex = text.indexOf("{br}");

        if (brIndex > 0) {
            let row = text.split("{br}");

            let maxBlockNum = 0;
            for (let b = 0; b < row.length; b++) {
                let chars = row[b].split("{space}");
                if (chars.length === 0) {
                    continue;
                }
                maxBlockNum = chars.length > maxBlockNum ? chars.length : maxBlockNum;
            }
            if (maxBlockNum === 0) {
                return;
            }


            let charLen = options.textWidth / maxBlockNum;
            for (let b = 0; b < row.length; b++) {
                let chars = row[b].split("{space}");
                if (chars.length === 0) {
                    continue;
                }
                for (let j = 0; j < chars.length; j++) {
                    let charStr = chars[j];
                    context.fillText(charStr, x + charLen * j, y + b * options.lineHeight);
                    if (options.fontStrokeWidth > 0) {
                        context.strokeText(charStr, x + charLen * j, y + b * (options.lineHeight));
                    }
                }
            }
        } else if (textWidth > options.textWidth) { // 多行
            let chr = text.split("");
            let temp = "";
            let row = [];
            for (let a = 0; a < text.length; a++) {
                if (context.measureText(temp).width < options.textWidth && context.measureText(temp + (chr[a])).width <= options.textWidth) {
                    temp += chr[a];
                }
                else {
                    row.push(temp);
                    temp = chr[a];
                }
            }
            row.push(temp);

            for (let b = 0; b < row.length; b++) {
                context.fillText(row[b], x, y + b * options.lineHeight);
                if (options.fontStrokeWidth > 0) {
                    context.strokeText(row[b], x, y + b * options.lineHeight);
                }
            }
        } else {
            context.fillText(text, x, y);
            if (options.fontStrokeWidth > 0) {
                context.strokeText(text, x, y);
            }
        }
    }

    //根据素材获取缩放后宽高, 缩放比例
    getEditorSetting(asset) {
        let width = asset.size[0] < this.options.maxEditorWidth ? asset.size[0] : this.options.maxEditorWidth;
        // let width = this.options.maxEditorWidth;
        let scale = width / asset.size[0];
        // 容器宽高
        let editorWidth = width;
        let editorHeight = scale * asset.size[1];
        // let editorHeight = this.options.containerHeight;

        return { width: editorWidth, height: editorHeight, scale: scale };
    }

    // 创建图片编辑控件
    createImageEditButton(id) {
        let EDITOR = this;
        if (EDITOR.widgets[id]) {
            EDITOR.widgets[id].node.show();
            return;
        } else {
            EDITOR.widgets[id] = true;
        }

        // 计算截图框大小
        let asset = this.currentAssets[id];
        let ui = asset.ui;
        let originWidth = asset.size[0];
        let originHeight = asset.size[1];

        // let setting = this.getEditorSetting(asset);
        let setting = this.getEditorInfo(asset);
        let scale = setting.scale;

        // 容器宽高
        let editorWidth = setting.width;
        let editorHeight = setting.height;

        asset.editorWidth = editorWidth;
        asset.editorHeight = editorHeight;

        asset.originWidth = originWidth;
        asset.originHeight = originHeight;

        asset.originScale = scale;
        asset.displayScale = asset.ui.s[0] * scale;

        let anchorStr = 0 + "px " + 0 + "px";

        // 缩放矩阵
        let preAffine = new Affine();
        preAffine.mA = scale;
        preAffine.mD = scale;


        // 原始矩阵
        let anchor = { x: ui.a[0], y: ui.a[1] };
        let position = { x: ui.p[0], y: ui.p[1] };
        let zoom = { x: ui.s[1], y: ui.s[1] };
        let rotation = ui.r;

        let origin = new Affine();
        origin.set(anchor, position, zoom, rotation);
        asset.inverse = origin.getInverse();
        origin.multiply(preAffine);

        let uIndex = asset.ui.index + 1; // FIXME

        let matrix = origin.mA + ", " + origin.mB + "," + origin.mC + ", " + origin.mD + "," + origin.mE + "," + origin.mF;
        let style = "width: " + asset.size[0] + "px"
            + "; height: " + asset.size[1] + "px"
            + "; z-index:" + uIndex
            + "; transform: matrix(" + matrix + ")"
            + "; -webkit-transform-origin:" + anchorStr;


        let option = {
            style: style,
            mask_img_src: ""
        };

        // 设置遮罩层
        if (asset.ui.f && asset.ui.f.length > 1) {
            option.mask_img_src = asset.ui.f;
        }

        let widget = this.buildImageEditButton(id, option);

        widget.matrix = origin;
        widget.moveData = {};
        EDITOR.widgets[id] = widget;
        widget.parentNode = this.container;

        widget.node.insertBefore(this.groupBoxes[widget.group_id].emptyNode);
        widget.maskNode.insertAfter(this.groupBoxes[widget.group_id].emptyNode);
        return widget;
    }

    // 截图保存数据
    doImageCrop(id, widget, callback) {
        let EDITOR = this;
        EDITOR.log('widget', widget);
        EDITOR.log("do crop : " + id + " cropper : " + typeof (widget.cropper));

        if (widget.cropper === undefined) return;
        let cropper = widget.cropper;

        let asset = EDITOR.currentAssets[id];
        let canvas = cropper.getCroppedCanvas({
            width: asset.size[0],
            height: asset.size[1],
            currentItem: asset,
        });

        if (canvas) {
            if (EDITOR.debug) {
                let img = new Image();
                img.src = canvas.toDataURL("image/png");
                $("body").append(img);
                img = null;
            }

            // 转成 blob
            canvas.toBlob(function (blob) {
                if (!EDITOR.croppedData[id]) {
                    EDITOR.croppedCount++;
                }
                delete EDITOR.croppedData[id];

                EDITOR.croppedData[id] = null;
                canvas = null;

                EDITOR.croppedData[id] = blob;
                EDITOR.log(EDITOR.croppedData);

                EDITOR.editedAssets[id] = true;

                if (callback) {
                    callback();
                }
            }, "image/png", 1);
        }
    }
    // 构建 图片截图控件
    createImageCropper(id, widget) {
        let EDITOR = this;
        let asset = this.currentAssets[id];

        let cropperOptions = {
            containerWidth: 200,
            containerHeight: 400,
            cropperWidth: 200,
            cropperHeight: 400,
            originWidth: asset.size[0],
            originHeight: asset.size[1],
            zIndex: asset.ui.index + 1, // FIXME
            asset: asset,
            matrix: widget.matrix,
            originScale: asset.originScale
        };

        cropperOptions.containerWidth = EDITOR.options.containerWidth;
        cropperOptions.containerHeight = EDITOR.options.containerHeight;
        cropperOptions.cropperWidth = asset.originWidth;
        cropperOptions.cropperHeight = asset.originHeight;

        if (widget.cropper) {
            widget.cropper = null;
        }

        cropperOptions.ready = function (cropper) {
            cropper.cropBox.insertBefore(EDITOR.groupBoxes[EDITOR.currentGroupId].emptyNode);
            widget.cropperBoxNode = cropper.cropBox;

            EDITOR.doImageCrop(id, widget);
        };

        let image_to_crop = widget.cropImgNode.get(0);
        widget.cropper = new VCropper(image_to_crop, cropperOptions, this);
        widget.moveData = widget.cropper.moveData;
        widget.node.hide();
    }

    // 构建 图片编辑控件
    buildImageEditButton(id, option) {
        let EDITOR = this;
        let widget = {};
        widget.id = id;
        widget.asset = this.currentAssets[id];
        widget.moveable = widget.asset.moveable;
        widget.editable = widget.asset.editable;
        widget.group_id = widget.asset.ui.group;
        widget.type = widget.asset.ui.type;

        let $img_box = $(this.htmls.img_box_html);
        widget.node = $img_box.attr("style", option.style).attr("widget-id", id);
        widget.imgButtonNode = $img_box.find(".ve-img-button").attr("id", id);//.attr("style", option.style);
        widget.inputNode = $img_box.find("input");
        widget.maskNode = $(this.htmls.mask_html);

        // 图片控件点击事件
        widget.click = function (e) {
            e.stopPropagation();
            if (!widget.editable) return;
            widget.inputNode.trigger('click');
        };

        widget.inputNode.on("click", function (e) {
            e.stopPropagation();
        });

        // 配饰图
        if (option.mask_img_src !== "") {
            let img = new Image();
            img.src = option.mask_img_src;
            img.onload = function () {
                widget.maskNode.css({
                    "background-image": "url(" + img.src + ") ",
                    "background-repeat": "no-repeat",
                    "z-index": widget.asset.ui.index,
                    "display": "block"
                }).show();
            };
        }

        // 是否可编辑
        if (widget.editable) {
            widget.inputNode.on("change", function (e) {

                let target = e.target;
                let done = function (url) {
                    let $crop_box = $(EDITOR.htmls.crop_html);
                    widget.cropImgNode = $($crop_box.find(".ve-img-to-crop").get(0));
                    widget.cropImgNode.attr("src", url);

                    // 删除旧图
                    if (widget.imgContainerNode) {
                        widget.imgContainerNode.remove();
                    }

                    widget.imgContainerNode = $crop_box;
                    widget.imgButtonNode.append($crop_box);

                    // 构建截图器
                    EDITOR.createImageCropper(id, widget);
                };

                // load local file
                let files = target.files;
                let reader;
                let file;
                if (files && files.length > 0) {
                    file = files[0];
                    if (URL) {
                        done(URL.createObjectURL(file));
                    } else if (FileReader) {
                        reader = new FileReader();
                        reader.readAsDataURL(file);
                        reader.onload = function () {
                            done(reader.result);
                        };
                    }
                }
            });

        }
        return widget;
    }

    // 初始化 html
    initHtml() {
        let html = {};
        html.img_box_html = `<div class="ve-img-box ve-widget"><div class="ve-img-button" ></div>  
            <input type="file" class="ve-select-image" name="image" accept="image/*" style="display: none" />
            </div>`;

        html.mask_html = '<div class="ve-widget-mask" style="display:none"></div>';
        html.crop_html = `<div class="ve-img-container"><img class="ve-img-to-crop" ></div>`;
        html.debug_box_html = `<div class="ve-debug-box" ><span class="ve-empty-line"></span></div>`;
        html.text_box_html = `<div class="ve-text-box ve-widget"><div class="ve-text-container">
                <textarea class="ve-text-input" name="ve-text" ></textarea>
            </div></div>`;
        this.htmls = html;
    }
}
