/**
 * file uploader based on qiniu js sdk 2.x
 *
 * @author dongyado
 * */

class VUploader {

    constructor(options) {
        this.options = options;

        // 上传　token
        this.uptoken = "";
        if (this.options.uptoken) {
            this.uptoken = this.options.uptoken;
        }

        // uptoken get api
        this.uptoken_url = "";
        if (this.options.uptoken_url) {
            this.uptoken_url = this.options.uptoken_url;
        }


        // c
        if (this.uptoken === "" && this.uptoken_url === "") {
            throw new Error("uptoken or uptoken_url must be set in options");
        }

        this.upload_config = {
            useCdnDomain: true,
        };

        if (this.options.upload_config) {
            this.upload_config = this.options.upload_config;
        }

        // 上传文件前缀
        this.file_prefix = "";
        if (this.options.file_prefix) {
            this.file_prefix = this.options.file_prefix;
        }

        // 文件后缀
        this.file_ext = "";
        if (this.options.file_ext) {
            this.file_ext = this.options.file_ext;
        }

        this.put_extra = {
            fname: "",
            params: {},
            mimeType: null
        };

        // init uploader
        this.init();
        this.observers     = {};
        this.subscriptions = {};
    }


    /**
     * 生成文件名
     *
     * */
    genFileKey(name) {
        let date     = new Date();
        let timestamp = date.getFullYear() + '/' + (date.getMonth() + 1) + '/' + date.getDate() + "/" + date.getTime();
        let filename = name.substr(0, name.indexOf("."));
        let ext      = name.substr(name.indexOf("."));
        let key      = this.file_prefix + timestamp + filename + ext;
        return key;
    }


    /**
     * 初始化上传器，提前获取　uptoken
     *
     * */
    init() {
        let UPLOADER = this;
        // 如果没有指定 uptoken，通过 uptoken_url 获取
        if (this.uptoken === "") {
            $.ajax({
                type: 'POST',
                url: UPLOADER.uptoken_url,
                data: {},
                dataType: "json",
                async: true,
                before: function (e) {
                },
                success: function (data) {
                    if (data.errno === 0) {
                        UPLOADER.uptoken = data.data.uptoken;
                    }
                },
                error: function (data) {
                    throw new Error("get uptoken failed width uptoken_url :  " + UPLOADER.uptoken_url);
                }
            });
        }
    }



    /**
     *  生成文件 唯一 uploadKey 值
     *
     * @param blob，文件数据
     * @param id, 文件唯一id, 用于计算完成后回调
     * @param callback, 计算完成回调
     */
    generatorUploadKey(blob, id, callback) {
        let reader = new FileReader();
        reader.readAsBinaryString(blob);
        reader.onloadend = function () {
            let hash = Math.random() * 10 + (new Date()).valueOf();
            callback(blob, hash, id);
        };
    }


    /**
     * 上传图片
     *
     * @param file, blob
     * @param filename, 文件名
     * @param genFilename, 是否生成文件名
     * @param fileExt, 指定文件后缀
     * @param handlers, 上传回调函数
     *
     * */
    upload(file, filename, genFilename, fileExt, handlers) {
        let UPLOADER                 = this;
        if (UPLOADER.uptoken === "") {
            throw new Error("invalid uptoken  ");
        }

        UPLOADER.observers[filename] = {
            next: function (res) {
                if (handlers.uploadNext) {
                    handlers.uploadNext(filename, res);
                }
            },
            error: function (err) {
                if (handlers.uploadError) {
                    handlers.uploadError(filename, err);
                }
            },
            complete: function (res) {
                if (handlers && handlers.uploadSuccess) {
                    handlers.uploadSuccess(filename, res);
                }
            }
        };

        // 生成文件名
        if (genFilename) {
            UPLOADER.generatorUploadKey(file, filename, function (file, hash, filename) {
                let file_ext                     = fileExt ? fileExt : UPLOADER.file_ext;
                let tmp_key                      = file_ext !== "" ? hash + "." + file_ext : hash;
                let key                          = UPLOADER.genFileKey(tmp_key);
                let observable                   = qiniu.upload(file, key, UPLOADER.uptoken.token, UPLOADER.put_extra, UPLOADER.upload_config);
                UPLOADER.subscriptions[filename] = observable.subscribe(UPLOADER.observers[filename]);
            });
        } else {
            let observable                   = qiniu.upload(file, filename, UPLOADER.uptoken.token, UPLOADER.put_extra, UPLOADER.upload_config);
            UPLOADER.subscriptions[filename] = observable.subscribe(UPLOADER.observers[filename]);
        }
    }
}
